import inspect

from .constants import SINGLETON, SIMPLE_TYPES
from .settings import Settings, EMPTY_SETTINGS
from .registry import Registry
from .types import Target


class Builder:
    _registry: Registry
    _settings: dict[type[Target], Settings]
    _cache: dict[type[Target], Target]
    _previous: 'Builder'
    _classes: set[type[Target]]

    def __init__(
        self,
        registry: Registry,
        settings: dict[type[Target], Settings],
        cache: dict[type[Target], Target],
        previous: 'Builder' = None,
    ):
        self._registry = registry
        self._settings = settings
        self._cache = cache
        self._previous = previous
        self._classes = set()

    def get_settings(self, target: Target) -> tuple[Settings, 'Builder']:
        """
        Ищет настройку в кеше, если не находит, рекурсивно обращается к предкам,
        ища настройку у них. Если не находит, возвращает пустые настройки.

        Возвращает настройку вместе с объектом-сборщиком,
        чтобы потом можно было поместить созданный инстанс в кеш того же слоя,
        с которого использовались настройки.
        """
        if cls_settings := self._settings.get(target):
            return cls_settings, self

        if self._previous:
            return self._previous.get_settings(target)
        else:
            return EMPTY_SETTINGS, self

    def get_cached(self, target: Target) -> Target | None:
        """
        Ищет инстанс для указанного класса в кеше.
        Если не находит, рекурсивно обращается к предкам,
        ища настройку у них.
        Возвращает None, если инстанс для класса не найден.
        """
        if cached := self._cache.get(target):
            return cached

        if self._previous:
            return self._previous.get_cached(target)
        else:
            return None

    def cache_instance(self, target: type[Target], instance: Target) -> None:
        """
        Помещает инстанс в кеш.
        """
        self._cache[target] = instance

    def detect_cycle(self, target: Target) -> None:
        """
        Проверка на цикл в графе зависимостей.

        Должно проверяться при каждом вызове build,
        для всех сборщиков одновременно. Для этого контейнер создает множество,
        которое затем передает каждому сборщику. Это проверка помещает
        указанный класс в множество.
        Если указанный класс встретится во множестве, значит, в графе есть цикл.
        """
        if target in self._classes:
            raise ValueError(f'Cycle reference detected on class {target}')

        self._classes.add(target)

    def build(self, target: type[Target]) -> Target:
        """
        Собирает объект указанного класса.

        Рекурсивно вызывается для каждого аргумент фабрики,
        найденной для указанного класса.
        """

        # Если объект уже есть в кеше, то можно просто его отдать
        if cached := self.get_cached(target):
            return cached

        # Ищем настройки для указанного класса
        target_settings, target_settings_layer = self.get_settings(target)

        # Инстанс, заданный в настройках возвращается как есть
        if target_settings.instance_:
            return target_settings.instance_

        self.detect_cycle(target)

        # Выбираем фабрику для указанного класса
        factory = target_settings.factory_ or self._registry.get(target)
        factory_settings = self.get_settings(factory)[0]

        # Фабрика выбрана, далее нужно собрать аргументы.
        # Нужно получаем сигнатуру для фабрики,
        # чтобы по ней построить аргументы для вызова фабрики
        factory_kwargs = {}
        for name, parameter in self._registry.signature(factory).items():

            # Если для параметра в init было указанно значение,
            # то берем значение "как есть".
            if name in factory_settings.init_:
                factory_kwargs[name] = factory_settings.init_[name]
                continue

            # Для аргументов простых типов контейнер не ищет фабрики
            if parameter.annotation in SIMPLE_TYPES:
                continue

            # Инстанцирование аргумента
            if inspect.isclass(parameter.annotation):
                instance = self.build(parameter.annotation)
                if instance is not None:
                    factory_kwargs[name] = instance

                # Случай, когда нечего указать в обязательный аргумент
                elif parameter.default is inspect.Parameter.empty:
                    raise ValueError(
                        f"Can't resole attribute {name} "
                        f"for {factory}, attribute don't have default value "
                        f"and {factory} has returned None. "
                        f"Maybe you have forgot to add 'return' "
                        f"to the end of your factory?"
                    )

                factory_kwargs[name] = instance

        # Постройка инстанса указанного класса
        instance = factory(**factory_kwargs)

        # TRANSIENT объекты не кешируются,
        # чтобы контейнер при каждом запросе строил их заново
        if instance and target_settings.scope_ == SINGLETON:
            target_settings_layer.cache_instance(target, instance)

        return instance



from enum import Enum, IntEnum
from datetime import date, datetime
from uuid import UUID


SINGLETON = 'SINGLETON'
TRANSIENT = 'TRANSIENT'

SCOPES = (SINGLETON, TRANSIENT)
SIMPLE_TYPES = (
    int, str, float,
    Enum, IntEnum,
    UUID,
    date, datetime
)



from abc import ABC
from collections import defaultdict
from functools import lru_cache
from itertools import chain
import inspect
from types import ModuleType
from typing import (
    Optional, Callable, get_type_hints, Tuple, Sequence, Generator, get_origin,
)

from .types import Factory, Registerable, ModuleOrTarget, Target


class Registry:
    """
    Хранит в себе информацию о классах, их фабриках и их зависимостях.

    Предполагается что этот объект инстанцируется контейнером
    и используется только "под капотом".
    """

    _storage = dict[Target, list[Factory[Target]]]

    def __init__(self):
        self._storage = defaultdict(list)
        self._signatures_cache = {}

    @lru_cache(1000)
    def signature(self, cls: Target) -> dict[str, inspect.Parameter]:
        """
        Возвращает описание сигнатуры указанной фабрики.
        Отличается от обычного inspect.signature тем,
        что пытается разрешить ForwardReference.

        Например:
        >>> class Test:
        ...     field: 'Test'
        """
        hints = get_type_hints(cls)
        signature = inspect.signature(cls)

        result = {}
        for name, parameter in signature.parameters.items():
            if isinstance(parameter.annotation, str):
                hint = hints[name]
            else:
                hint = parameter.annotation

            result[name] = parameter.replace(annotation=hint)

        return result

    def get(self, target: Target) -> Factory[Target]:
        """
        Вернет фабрику для указанного класса.

        Выбросит исключение, если фабрики не нашлось, или ее больше одной.
        """
        factories = self._storage.get(target)

        if not factories:
            raise ValueError(
                f'Class {target} do not have registered implementations.\n'
            )

        if len(factories) > 1:
            raise ValueError(
                f'Can not to resolve {target}, '
                f'implementations are: {factories}\n'
            )

        return factories[0]

    def register(self, *targets: ModuleOrTarget) -> None:
        """
        Запускает регистрацию указанного объекта.
        Выбрасывает ValueError, если указанный объект
        не является классом, функцией или модулем.
        """
        for target in targets:
            if self._register(target) is None:
                raise ValueError(
                    f'Registration target must be class, '
                    f'function or module, {target} is {type(target)}'
                )

    def _register(self, target: ModuleOrTarget) -> Optional[str]:
        """
        Идентификация объекта и регистрация в реестре.
        """
        result = None

        if inspect.ismodule(target):
            self._register_module(target)
            result = 'module'

        elif inspect.isabstract(target):
            self._register_interface(target)
            result = 'interface'

        elif inspect.isfunction(target):
            self._register_function(target)
            result = 'function'

        elif inspect.isclass(target):
            self._register_class(target)
            result = 'class'

        elif _is_generic(target):
            self._register_interface(target)
            result = 'generic'

        return result

    def _register_interface(self, interface: ABC) -> None:
        """
        Регистрация интерфейса/абстрактного класса.
        ABC считаются классами без фабрики.
        """
        self._add_entry(interface)

    def _register_class(self, cls: type[object]) -> None:
        """
        Регистрация обычного класса.
        Обычный класс считается фабрикой для самого себя,
        и для каждого предка в MRO.
        """
        self._add_entry(cls, cls)
        for ancestor in _get_interfaces_for_cls(cls):
            self._add_entry(ancestor, cls)

    def _register_function(self, func: Callable) -> None:
        """
        Регистрация функции.
        Если в аннотации функции указано, что она возвращает инстанс
        какого-либо класса, то она считается фабрикой для этого класса.

        Остальные игнорируются.
        """
        factory_returns = _get_factory_result(func)
        if not inspect.isclass(factory_returns):
            return

        self._add_entry(factory_returns, func)

    def _register_module(self, module: ModuleType) -> None:
        """
        Регистрация модуля.

        Регистрируется содержимое модуля, вместе с дочерними модулями,
        рекурсивно, но не содержимое сторонних модулей.

        Ради примера представим себе содержимое __init__ у модуля example:
        >>> import functools
        ... from itertools import chain
        ... from . import child  # NOQA

        При регистрации:
        >>> from classic.container import container
        ... container.register(child)

        В реестр попадут example, example.child и chain, но не functools.
        """
        targets, submodules = _get_members(module)
        for target in chain(targets, submodules):
            self._register(target)

    def _add_entry(
        self, cls: Registerable,
        factory: Optional[Factory[Target]] = None,
    ) -> None:
        """
        Добавление записи в реестр.
        """
        factories = self._storage[cls]
        if factory is not None and factory not in factories:
            factories.append(factory)


def _is_submodule(submodule: ModuleType, module: ModuleType) -> bool:
    return submodule.__name__.startswith(module.__name__)


def _is_generic(target: type[object]) -> bool:
    return get_origin(target) is not None


def _get_members(module: ModuleType) -> Tuple[Sequence[Registerable],
                                              Sequence[ModuleType]]:
    submodules = []
    targets = []

    for name, member in module.__dict__.items():
        if name.startswith('_'):
            continue
        if isinstance(member, ModuleType) and _is_submodule(member, module):
            submodules.append(member)
        elif inspect.isclass(member):
            targets.append(member)

    return targets, submodules


def _get_interfaces_for_cls(
    target: type[Target],
) -> Generator[type[object], None, None]:

    for cls in target.__mro__:
        if cls != object:
            yield cls


def _get_factory_result(factory: Factory[Target]) -> type[Target] | None:
    if inspect.isclass(factory):
        return factory

    signature = inspect.signature(factory)
    result = signature.return_annotation
    if result == inspect.Parameter.empty:
        return None

    return result



from classic.container import factory, instance

from example import (
    Composition, Interface, Implementation1, Implementation2,
    NextLevelComposition,
)


def test_nested_resolve(container):
    container.register(
        Composition, Interface,
        Implementation1, Implementation2,
    )

    def make_obj() -> Interface:
        return container.resolve(Implementation1)

    container.add_settings({
        Interface: factory(make_obj),
    })

    obj = container.resolve(Composition)

    assert isinstance(obj, Composition)
    assert hasattr(obj, 'impl')
    assert isinstance(obj.impl, Implementation1)


def test_nested_resolve_with_settings(container):
    container.register(
        Composition, Interface,
        Implementation1, Implementation2,
    )

    def make_obj() -> Interface:
        return container.resolve(
            Interface, {Interface: factory(Implementation1)},
        )

    container.add_settings({
        Interface: factory(make_obj),
    })

    obj = container.resolve(Composition)

    assert isinstance(obj, Composition)
    assert hasattr(obj, 'impl')
    assert isinstance(obj.impl, Implementation1)


def test_nested_resolve_with_instance(container):
    container.register(
        Composition, Interface,
        Implementation1, Implementation2,
    )

    just_obj = object()

    def make_obj() -> Interface:
        return container.resolve(
            Interface, {Interface: instance(just_obj)},
        )

    container.add_settings({
        Interface: factory(make_obj),
    })

    obj = container.resolve(Composition)

    assert isinstance(obj, Composition)
    assert hasattr(obj, 'impl')
    assert obj.impl is just_obj


def test_nested_resolve_with_simple_instancing(container):
    container.register(
        Composition, Interface,
        Implementation1, Implementation2,
        NextLevelComposition,
    )

    def make_obj() -> Composition:
        return Composition(
            impl=container.resolve(Implementation1),
        )

    container.add_settings({
        Composition: factory(make_obj),
    })

    obj = container.resolve(NextLevelComposition)

    assert isinstance(obj, NextLevelComposition)
    assert isinstance(obj.obj, Composition)
    assert isinstance(obj.obj.impl, Implementation1)



import pytest
from classic.container import Settings, TRANSIENT

from example import Implementation1


@pytest.fixture
def settings():
    return Settings()


def test_chain_responsibility_factory(settings):
    factory_settings = settings.factory(Implementation1)
    assert settings is factory_settings


def test_chain_responsibility_init(settings):
    init_settings = settings.init(some=1)
    assert settings is init_settings


def test_chain_responsibility_scope(settings):
    scope_settings = settings.scope(name=TRANSIENT)
    assert settings is scope_settings


def test_chain_responsibility_instance(settings):
    impl = Implementation1()
    impl_settings = settings.instance(instance=impl)
    assert settings is impl_settings


def test_chain_of_responsibility(settings):
    many_settings = settings.factory(
        Implementation1
    ).init(some=1).scope(name=TRANSIENT)
    assert settings is many_settings


def test_setting_raise_error(container):

    with pytest.raises(AssertionError):
        container.add_settings({
            Implementation1: Settings(
                instance=Implementation1(), factory=Implementation1
            ),
        })

    with pytest.raises(AssertionError):
        container.add_settings({
            Implementation1: Settings().instance(Implementation1()).factory(
                Implementation1
            ),
        })

    with pytest.raises(AssertionError):
        container.add_settings({
            Implementation1: Settings().instance(
                Implementation1()
            ).init(some=1),
        })

    with pytest.raises(AssertionError):
        container.add_settings({
            Implementation1: Settings().instance(Implementation1()).scope(
                TRANSIENT
            ),
        })

    with pytest.raises(AssertionError):
        container.add_settings({
            Implementation1: Settings().factory(Implementation1).instance(
                Implementation1()
            ).scope(TRANSIENT),
        })



from abc import ABC
from types import ModuleType
from typing import TypeAlias, Callable, TypeVar


Target = TypeVar('Target', bound=object)
ModuleOrTarget: TypeAlias = Target | ModuleType
Factory: TypeAlias = type[Target] | Callable[[...], Target]
Registerable: TypeAlias = Factory | ABC



import threading
from typing import TypeVar, Optional

from .registry import Registry
from .builder import Builder
from .settings import Settings


Object = TypeVar('Object', bound=object)


class Container:
    """
    Классический IoC-контейнер.

    """

    _registry: Registry
    _settings: dict[type[Object], Object]
    _cache: dict[type[Object], Object]
    _lock: threading.RLock
    _current_builder: Optional[Builder]

    def __init__(self):
        self._lock = threading.RLock()
        self._registry = Registry()
        self._settings = {}
        self._cache = {}
        self._current_builder = None

    def register(self, *args: object) -> None:
        """
        При обращении принимает список компонентов.
        Определяет тип, заносит в реестр, каждому типу сопоставляет
        список фабрик, способных построить указанный тип.

        Элементами списка могут быть: абстрактные классы, классы,
        фабрики (функции, возвращающие один инстанс любого класса) и модули.

        - Абстрактные классы регистрируется в реестре только как ключи,
          без фабрик.
        - Нормальные классы регистрируются как ключ и соответсвующая ему
          фабрика - конструктор самого класса.
        - Фабрики регистрируются сложнее, ключом будет являться результат из
          аннотации функции, а значением сама фабрика.
          Пример:
          >>> def some_factory() -> SomeClass:  # NOQA
          ...     # будет зарегистрировано как SomeClass: [some_factory]
          ...     pass

        - Модули не регистрируются напрямую. Регистратор рекурсивно обходит
          указанный модуль и все его дочерние модули, и регистрирует в реестре
          все классы и фабрики из каждого модуля.
          >>> import os
          ... # будут зарегистрированы os и os.path
          ... # но не sys
          ... container.register(os)  # NOQA

        """
        with self._lock:
            self._registry.register(*args)

    def resolve(
        self, target: type[Object],
        settings: dict[type[object], Settings] = None,
    ) -> Object:
        """
        Собирает объект с его деревом зависимостей для указанной реализации.

        Рекурсивно обходит дерево зависимостей, начиная с указанного класса.
        На каждый шаг рекурсии для указанного класса ищется фабрика в реестре.
        Далее для найденной фабрики собираются аргументы,
        чтобы вызвать фабрику и построить объект.
        При этом:
         - пропускаются аргументы простых типов, аргументы без аннотаций
           и функции;
         - подставляются значения из init для аргументов,
           указанных в этом же init;
         - для аргументов, проаннотированных классами, повторяется рекурсия.

        В процессе разрешения могут возникать ситуации, когда:
         - для интерфейса (абстрактного класса) не нашлось реализации;
         - для класса нашлось больше 1 фабрики
           и ни одна не указана в настройках для этого класса;
         - фабрика для аргумента вернула None
           и для аргумента не указан значение по умолчанию;
         - при вызове фабрики не был указан обязательный аргумент.
        Во всех этих случаях контейнер выкидывает ValueError с описанием ошибки.
        """

        with self._lock:

            # Ссылку на предыдущий сборщик нужно запомнить,
            # чтобы после завершения resolve можно было
            # восстановить ссылку в _current_resolve
            previous = self._current_builder

            self._current_builder = Builder(
                registry=self._registry,
                settings=self._settings if not previous else settings or {},
                cache=self._cache if not previous else {},
                previous=previous,
            )
            result = self._current_builder.build(target)

            self._current_builder = previous

        return result

    def add_settings(self, settings: dict[type[object], Settings]) -> None:
        """
        Добавляет или обновляет настройки контейнера.

        Принимает словарь, в котором ключи - классы,
        значения - объекты Settings.
        """
        with self._lock:
            self._settings.update(settings)

            # Логично предположить, что если для классов указано что-либо
            # в настройках, то можно это автоматически регистрировать
            self.register(*settings.keys())

    def reset(self):
        """
        Удаляет добавленные настройки контейнера и ссылки на инстансы ранее
        созданных классов. Не должен вызываться во время resolve.
        """

        # В случае, если reset запрошен во время resolve,
        # дешевле выбросить исключение, чем разгребать
        # очень странные ошибки
        assert self._current_builder is None

        with self._lock:
            self._settings.clear()
            self._cache.clear()



import pytest

from classic.container import TRANSIENT, factory, scope, instance, init

import example
from example import (
    AnotherCls, Implementation1, Implementation2, Interface,
    SomeCls, YetAnotherCls, Composition, composition_factory,
    NextLevelComposition, empty_factory, some_func,
    SelfReferenced, CycledA, CycledB, SomeGeneric,
)


@pytest.mark.parametrize('obj', [
    (1,),
    (Implementation1(),),
])
def test_registration_error(obj, container):

    with pytest.raises(ValueError):
        container.register(obj)


def check_class_resolve(container):

    resolved = container.resolve(YetAnotherCls)

    assert resolved is not None
    assert isinstance(resolved, YetAnotherCls)
    assert isinstance(resolved.some, SomeCls)
    assert isinstance(resolved.another, AnotherCls)
    assert isinstance(resolved.another.some, SomeCls)
    assert resolved.some is resolved.another.some


def check_resolve_impl(container):
    resolved = container.resolve(Interface)

    assert isinstance(resolved, Implementation1)


def check_next_level_composition_resolve(container):
    resolved = container.resolve(NextLevelComposition)

    assert isinstance(resolved, NextLevelComposition)
    assert isinstance(resolved.obj, Composition)
    assert isinstance(resolved.obj.impl, Implementation1)


def test_simple_resolving(container):
    container.register(SomeCls, AnotherCls, YetAnotherCls)

    check_class_resolve(container)


def test_resolve_with_abc(container):
    container.register(Interface, Implementation1)

    check_resolve_impl(container)


def test_resolve_without_implementation(container):
    container.register(Interface)

    with pytest.raises(ValueError):
        container.resolve(Interface)


def test_resolve_without_registration(container):
    with pytest.raises(ValueError):
        container.resolve(Interface)


def test_resolve_with_type_error(container):
    container.add_settings({Interface: factory(some_func)})
    with pytest.raises(ValueError):
        container.resolve(Interface)


def test_resolve_empty_factory(container):
    container.register(Composition)
    container.add_settings({
        Interface: factory(empty_factory)
    })
    with pytest.raises(ValueError):
        container.resolve(Composition)


def test_resolve_with_abc_implicit(container):
    container.register(Implementation1)

    check_resolve_impl(container)


def test_resolve_module_1(container):
    container.register(example)

    container.add_settings({
        Interface: factory(Implementation1),
        Composition: factory(composition_factory)
    })

    check_next_level_composition_resolve(container)


def test_resolve_module_2(container):
    container.register(example)

    check_class_resolve(container=container)


def test_resolve_instance(container):
    implementation = Implementation1()

    container.add_settings({
        Implementation1: instance(implementation),
    })
    resolved = container.resolve(Implementation1)

    assert resolved is implementation


def test_factory_calling(container):
    container.register(Implementation1, composition_factory)

    container.add_settings({
        Composition: factory(composition_factory),
    })

    resolved = container.resolve(Composition)

    assert isinstance(resolved, Composition)
    assert isinstance(resolved.impl, Implementation1)


def test_factory_is_create_objects(container):
    container.register(Implementation1, composition_factory,
                       Composition, NextLevelComposition)
    container.add_settings({
        Composition: factory(composition_factory)
    })

    check_next_level_composition_resolve(container)


def test_resolve_init(container):
    impl = Implementation1()

    container.add_settings({
        Composition: init(impl=impl),
    })
    resolved = container.resolve(Composition)

    assert isinstance(resolved, Composition)
    assert resolved.impl is impl


def test_raise_when_many_implementations(container):
    container.register(Interface, Implementation1, Implementation2)

    with pytest.raises(ValueError):
        container.resolve(Interface)


def test_resolve_with_replacement(container):
    container.register(Interface, Implementation1, Implementation2)

    container.add_settings({
        Interface: factory(Implementation1),
    })

    check_resolve_impl(container)


def test_singleton_scope(container):
    container.register(Implementation1)

    resolved_1 = container.resolve(Implementation1)
    resolved_2 = container.resolve(Implementation1)

    assert resolved_1 is resolved_2


def test_transient_scope(container):
    container.register(Implementation1)

    container.add_settings({
        Implementation1: scope(TRANSIENT)
    })

    resolved_1 = container.resolve(Implementation1)
    resolved_2 = container.resolve(Implementation1)

    assert resolved_1 is not resolved_2


def test_reset_resolved_instances(container):
    container.register(example)

    # Проверка разрешения зависимостей без ошибок до reset-a.
    container.add_settings({Interface: factory(Implementation1)})
    container.resolve(Interface)

    result_1 = container.resolve(YetAnotherCls)
    container.reset()
    result_2 = container.resolve(YetAnotherCls)

    assert result_1 is not result_2
    with pytest.raises(ValueError):
        container.resolve(Interface)


def test_cycle_detect(container):
    container.register(SelfReferenced, CycledA, CycledB)

    with pytest.raises(ValueError):
        container.resolve(SelfReferenced)

    with pytest.raises(ValueError):
        container.resolve(CycledA)

    with pytest.raises(ValueError):
        container.resolve(CycledB)


def test_generic(container):
    container.register(SomeGeneric, SomeGeneric[int])

    container.add_settings({
        SomeGeneric: factory(lambda: '123'),
        SomeGeneric[int]: factory(lambda: 10),
    })

    assert container.resolve(SomeGeneric) is '123'
    assert container.resolve(SomeGeneric[int]) is 10



from .constants import SINGLETON, TRANSIENT
from .settings import Settings, settings, init, scope, factory, instance
from .container import Container


# Дефолтный контейнер, сделан ради удобства,
# чтобы не инстанцировать каждый раз вручную.
# Конечно, вы можете создавать свои инстансы контейнера, когда вам нужно.
container = Container()



from .constants import SINGLETON, SCOPES
from .types import Factory


class Settings:
    """
    Класс хранит настройки resolv-а.

    Используется для хранения способа создания объекта или самого объекта
    при разрешении зависимостей.
    """
    def __init__(
        self, init: dict[str, object] = None, factory: Factory = None,
        scope: str = None, instance: object = None,
    ):
        assert scope is None or scope in SCOPES, \
            f'Scope name must be SINGLETON or TRANSIENT. Current is "{scope}"'

        assert instance is None or (
            instance is not None and
            not factory and
            not init and
            (scope is None or scope == 'SINGLETONE')
        ), (f'Container can use only instance or '
            f'factory, init and scope must be SINGLETON')

        self.scope_ = scope or SINGLETON
        self.init_ = init or {}
        self.factory_ = factory
        self.instance_ = instance

    def __repr__(self):
        rows = []
        if self.scope_:
            rows.append(f'scope={self.scope_}')
        if self.init_:
            rows.append(f'init={self.init_}')
        if self.factory_:
            rows.append(f'factory={self.factory_}')
        if self.instance_:
            rows.append(f'instance={self.instance_}')
        args = ', '.join(rows)
        return f'<container.Settings({args})>'

    def init(self, **kwargs: object) -> 'Settings':
        """
        Позволяет установить значения аргументов для фабрики
        при построении объекта. Самое частое использование - передача
        простых объектов (чисел, строк).

        >>> from classic.container import Container, Settings, init
        ...
        ... class SomeClass:
        ...     def __init__(self, some_value: int):
        ...         # Для int будет неудобно указывать фабрику,
        ...         # так как много у каких классов может быть параметр типа int
        ...         # (справедливо для любого простого типа), поэтому
        ...         # библиотека оставляет возможность
        ...         # указать параметр через init
        ...         self.some_value = some_value
        ...
        ... container = Container()
        ...
        ... container.register(SomeClass)
        ...
        ... # Длинный способ через конструктор
        ... container.add_settings({
        ...     SomeClass: Settings(init=dict(some_value=2))
        ... })
        ...
        ... # Вызов "цепочкой"
        ... container.add_settings({
        ...     SomeClass: Settings().init(some_value=2)
        ... })
        ...
        ... # А можно через алиас
        ... container.add_settings({SomeClass: init(some_value=2)})
        """

        assert self.instance_ is None, (
            f'Container can use only instance or init'
        )
        self.init_ = kwargs
        return self

    def factory(self, factory: Factory) -> 'Settings':
        """
        Позволяет явно передать способ создания компонента системы.
        Значением может являться любой вызываемый объект возвращающий
        инстанс любого объекта, и имеющий аннотацию типов.

        >>> from abc import ABC, abstractmethod
        ... from classic.container import Container, Settings, factory
        ...
        ... class Interface(ABC):
        ...
        ...     @abstractmethod
        ...     def method(self): ...
        ...
        ... class Implementation(Interface):
        ...
        ...     def method(self):
        ...         return 1
        ...
        ... class SomeClass:
        ...
        ...     def __init__(self, impl: Interface):
        ...         self.impl = impl
        ...
        ... def composition_factory(obj: Interface) -> SomeClass:
        ...     return SomeClass(obj)
        ...
        ... container = Container()
        ...
        ... container.register(Implementation, SomeClass, composition_factory)
        ...
        ... # Длинный способ через конструктор
        ... container.add_settings({
        ...     SomeClass: Settings(factory=composition_factory)
        ... })
        ...
        ... # Вызов "цепочкой"
        ... container.add_settings({
        ...     SomeClass: Settings().factory(factory=composition_factory)
        ... })
        ...
        ... # А можно через алиас
        ... container.add_settings({SomeClass: factory(composition_factory)})
        """

        assert self.instance_ is None, (
            f'Container can use only instance or factory'
        )
        self.factory_ = factory
        return self

    def scope(self, name: str) -> 'Settings':
        """
        Данная настройка регулирует жизненны цикл объекта, который может быть
        SINGLETON и TRANSIENT.

        При значении SINGLETON контейнер создаст объект только один раз,
        все последующие запросы будут использовать тот же самый объект.
        Является значением по умолчанию.

        При TRANSIENT контейнер будет создавать новый объект при каждом resolve.

        Для каждого класса настройка scope добавляется отдельно!

        >>> from abc import ABC, abstractmethod
        ... from classic.container import Container, Settings, TRANSIENT, scope
        ...
        ... class Interface(ABC):
        ...
        ...     @abstractmethod
        ...     def method(self): ...
        ...
        ... class Implementation(Interface):
        ...
        ...     def method(self):
        ...         return 1
        ...
        ... container = Container()
        ...
        ... container.register(Interface, Implementation)
        ...
        ... # Длинный способ через конструктор
        ... container.add_settings({Implementation: Settings(scope=TRANSIENT)})
        ...
        ... # Вызов "цепочкой"
        ... container.add_settings({
        ...     Implementation: Settings().scope(name=TRANSIENT)
        ... })
        ...
        ... # А можно через алиас
        ... container.add_settings({Implementation: scope(TRANSIENT)})
        """
        assert name in SCOPES, (
            f'Scope name must be SINGLETON or TRANSIENT. Current is "{name}"'
        )
        assert self.instance_ is None or name == 'SINGLETON', (
            f'Scope name must be SINGLETON then used instance'
        )
        self.scope_ = name
        return self

    def instance(self, instance: object) -> 'Settings':
        """
        Настройка позволяет подать готовый инстанс класса.

        Подразумевается основное использование при потребности подачи
        в разные классы готовых объектов, но настроенных по-разному.

        Класс сделан для удобства, тоже самое можно сделать через фабрики.

        >>> from abc import ABC, abstractmethod
        ... from classic.container import Container, Settings, instance
        ...
        ... class Interface(ABC):
        ...     some_value: int
        ...
        ...     @abstractmethod
        ...     def method(self): ...
        ...
        ... class Implementation(Interface):
        ...
        ...     def __init__(self, some_value):
        ...         self.some_value = some_value
        ...
        ...     def method(self):
        ...         return 1
        ...
        ... class SomeClass:
        ...
        ...     def __init__(self, impl: Interface):
        ...         self.impl = impl
        ...
        ... container = Container()
        ...
        ... container.register(
        ...     Interface, Implementation, SomeClass,
        ... )
        ...
        ... impl = Implementation(1)
        ...
        ... # Длинный способ через конструктор
        ... container.add_settings({
        ...     SomeClass: Settings(instance=impl)
        ... })
        ...
        ... # Вызов "цепочкой"
        ... container.add_settings({
        ...     SomeClass: Settings().instance(instance=impl)
        ... })
        ...
        ... # А можно через алиас
        ... container.add_settings({SomeClass: instance(impl)})
        """
        assert (
            not self.init_ and
            not self.factory_ and
            self.scope_ == 'SINGLETON'
        ), (f'Container can use only instance or '
            f'factory, init and scope must be SINGLETON')
        self.instance_ = instance
        return self


settings = Settings
EMPTY_SETTINGS = Settings()


def init(**kwargs: object) -> Settings:
    """
    Обертка для создания настроек компонентов с параметрами.
    Самое частое использование - передача простых объектов (чисел, строк).
    """
    return Settings(init=kwargs)


def factory(factory: Factory) -> Settings:
    """
    Обертка для создания настроек компонентов со способом получения объекта.
    Принимает способ создания объекта (фабрика, класс, абстрактный класс).
    """
    return Settings(factory=factory)


def scope(name: str) -> Settings:
    """
    Обертка для указания настройки scope у элемента приложения.
    Возможные значения указанны в константах текущего пакета:
    SINGLETON, TRANSIENT
    """
    return Settings(scope=name)


def instance(obj: object) -> Settings:
    """
    Обертка для создания настроек с готовым объектом при разрешении
    зависимостей.
    """
    return Settings(instance=obj)
